"""
Provide static plugin management.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, final, override

from betty.plugin import PluginDefinition, ResolvableDefinition, resolve_definition
from betty.plugin.error import PluginNotFound
from betty.plugin.repository import PluginRepository

if TYPE_CHECKING:
    from collections.abc import Iterator

    from betty.machine_name import ResolvableMachineName


@final
class StaticPluginRepository[PluginDefinitionT: PluginDefinition](
    PluginRepository[PluginDefinitionT]
):
    """
    A repository that is given a static collection of plugins, and exposes those.
    """

    def __init__(
        self,
        plugin_type: type[PluginDefinitionT],  # noqa: A002
        *plugins: ResolvableDefinition[PluginDefinitionT],
    ):
        super().__init__(plugin_type)
        self._plugins = {
            plugin.id: plugin
            for plugin in (resolve_definition(plugin) for plugin in plugins)
        }

    @override
    def get(self, plugin_id: ResolvableMachineName, /) -> PluginDefinitionT:
        try:
            return self._plugins[plugin_id]
        except KeyError:
            raise PluginNotFound(self.type.type(), plugin_id, list(self)) from None

    @override
    def __iter__(self) -> Iterator[PluginDefinitionT]:
        yield from self._plugins.values()



"""
Rich integration.
"""

from rich.style import Style
from rich.theme import Theme as RichTheme


class Theme(RichTheme):
    """
    A Rich theme for Betty's console.
    """

    def __init__(self):
        super().__init__(
            {
                "prompt": Style(color="bright_magenta", bold=True),
            }
        )



"""
Integrate Betty with `Webpack <https://webpack.js.org/>`_.

This module is internal.
"""

from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Self, final, override

from betty.document import DocumentProvider, DocumentVars
from betty.extension import Extension, ExtensionDefinition
from betty.extension.webpack import build
from betty.extension.webpack.build import EntryPointProvider
from betty.extension.webpack.jinja.filter import FILTERS
from betty.extension.webpack.jobs import _GenerateAssets
from betty.html import CssProvider, JsProvider
from betty.jinja import Filters, JinjaProvider
from betty.project import Project
from betty.project.generate import Generator
from betty.service.factory import Manufacturable
from betty.service.provider import service
from betty.service.requirement.project import require_project

if TYPE_CHECKING:
    from collections.abc import Sequence

    from betty.job.scheduler import Scheduler


@final
@ExtensionDefinition(
    "webpack",
    label="Webpack",
    assets_directory=Path(__file__).parent / "assets",
)
class Webpack(
    Generator,
    Extension[Project],
    CssProvider,
    JsProvider,
    JinjaProvider,
    DocumentProvider,
    Manufacturable,
):
    """
    .. plugin:: extension:webpack.
    """

    def __init__(self, *, project: Project):
        super().__init__()
        self._project = project

    @override
    @classmethod
    @require_project
    async def new(cls, project: Project, /) -> Self:
        return cls(project=project)

    @override
    async def generate(self, scheduler: Scheduler) -> None:

        await scheduler.add(
            _GenerateAssets(
                builder=await self.builder,
                cache_directory=self._project.app.binary_file_cache.with_scope(
                    "webpack"
                ).path,
                www_directory=self._project.www_directory,
            )
        )

    @override
    async def get_public_css_paths(self) -> Sequence[str]:
        return (
            "betty-static:///css/webpack/webpack-vendor.css",
            *(
                f"betty-static:///css/webpack/{entry_point.plugin().id}.css"
                for entry_point in await self._project_entry_point_providers()
                if (
                    entry_point.webpack_entry_point_directory_path() / "main.scss"
                ).is_file()
            ),
        )

    @override
    async def get_public_js_paths(self) -> Sequence[str]:
        return ("betty-static:///js/webpack-entry-loader.js",)

    @override
    def new_document_vars(self) -> DocumentVars:
        return {
            "webpack_js_entry_points": set(),
        }

    @override
    @property
    def filters(self) -> Filters:
        return FILTERS

    async def _project_entry_point_providers(
        self,
    ) -> Sequence[EntryPointProvider]:
        extensions = await self._project.extensions
        return [
            extension
            for extension in extensions.flatten()
            if isinstance(extension, EntryPointProvider)
        ]

    @service
    async def builder(self) -> build.Builder:
        """
        The Webpack builder.
        """
        return build.Builder(
            await self._project_entry_point_providers(),
            self._project.configuration.debug,
            await self._project.jinja,
            self._project.configuration.root_path,
            user=self._project.app.user,
        )



"""Provide Betty's data model API."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Self, final, override
from uuid import uuid4

from betty.definition.human_facing import CountableHumanFacingDefinition
from betty.hashid import hashid
from betty.json.linked_data import (
    JsonLdObject,
    LinkedDataDumpableWithSchemaJsonLdObject,
)
from betty.json.schema import JsonSchemaReference, String
from betty.locale.localizable.gettext import _, ngettext
from betty.locale.localize import DEFAULT_LOCALIZER
from betty.plugin import Plugin, PluginDefinition, PluginTypeDefinition
from betty.plugin.discovery.entry_point import EntryPointDiscovery
from betty.string import kebab_case_to_lower_camel_case

if TYPE_CHECKING:
    import builtins

    from betty.locale.localizable import (
        CountableLocalizable,
        Localizable,
        ResolvableLocalizable,
    )
    from betty.machine_name import ResolvableMachineName
    from betty.portable import PortableMapping
    from betty.project import Project


class NonPersistentId(str):
    """
    A randomly generated ID that is not persistent.

    Entities must have IDs for identification. However, not all entities can be provided with an ID that exists in the
    original data set (such as a third-party family tree loaded into Betty).

    Non-persistent IDs are helpful in case there is no external ID that can be used. However, as they do not persist
    when reloading an ancestry, they *MUST NOT* be in contexts where persistent identifiers are expected, such as in
    URLs.
    """

    __slots__ = ()

    def __new__(cls, entity_id: str | None = None, /):  # noqa: D102
        return super().__new__(cls, entity_id or str(uuid4()))


class Entity(LinkedDataDumpableWithSchemaJsonLdObject, Plugin["EntityDefinition"]):
    """
    An entity is a uniquely identifiable data container.

    To test your own subclasses, use :py:class:`betty.test_utils.model.EntityTestBase`.
    """

    def __init__(
        self,
        id: str | None = None,  # noqa: A002
        *args: Any,
        **kwargs: Any,
    ):
        self._id = NonPersistentId() if id is None else id
        self._public_id = self._id if id is None else hashid(id)
        super().__init__(*args, **kwargs)

    @override
    def __hash__(self) -> int:
        return hash(self.ancestry_id)

    @property
    def id(self) -> str:
        """
        The entity ID.

        This MUST be unique per entity type, per ancestry.
        """
        return self._id

    @property
    def public_id(self) -> str:
        """
        The public entity ID.

        This MUST be unique per entity type, per ancestry.

        A public ID consists of alphanumeric characters only, and can therefore safely be used across file systems.
        """
        return self._public_id

    @property
    def ancestry_id(self) -> tuple[builtins.type[Self], str]:
        """
        The ancestry ID.

        This MUST be unique per ancestry.
        """
        return type(self), self.id

    @property
    def label(self) -> Localizable:
        """
        The entity's human-readable label.
        """
        return _("{entity_type} {entity_id}").format(
            entity_type=self.plugin().label, entity_id=self.id
        )

    @override
    async def dump_linked_data(self, project: Project, /) -> PortableMapping:
        portable = await super().dump_linked_data(project)

        if persistent_id(self) and self.plugin().public_facing:
            url_generator = await project.url_generator
            portable["@id"] = url_generator.generate(
                f"betty-static:///{self.plugin().id}/{self.id}/index.json",
                absolute=True,
            )
        portable["id"] = self.id

        return portable

    @override
    @classmethod
    async def linked_data_schema(cls, project: Project, /) -> JsonLdObject:
        schema = await super().linked_data_schema(project)
        schema._def_name = f"{kebab_case_to_lower_camel_case(cls.plugin().id)}Entity"
        schema.title = cls.plugin().label.localize(DEFAULT_LOCALIZER)
        schema.add_property("$schema", JsonSchemaReference())
        schema.add_property("id", String(title="Entity ID"), False)

        return schema


@final
@PluginTypeDefinition(
    "entity",
    label=_("Entity"),
    label_plural=_("Entities"),
    label_countable=ngettext("{count} entity", "{count} entities"),
    description=_(
        "Entities represent the information in your ancestry, such as people and places."
    ),
    discovery=[EntryPointDiscovery("betty.entity_type")],
)
class EntityDefinition(CountableHumanFacingDefinition, PluginDefinition[Entity]):
    """
    .. plugin_type:: entity.
    """

    def __init__(
        self,
        plugin_id: ResolvableMachineName,
        *,
        label: ResolvableLocalizable,
        label_plural: ResolvableLocalizable,
        label_countable: CountableLocalizable,
        description: ResolvableLocalizable | None = None,
        public_facing: bool = True,
    ):
        super().__init__(
            plugin_id,
            label=label,
            label_plural=label_plural,
            label_countable=label_countable,
            description=description,
        )
        self._public_facing = public_facing

    @property
    def public_facing(self) -> bool:
        """
        Whether entities of this type are public-facing.
        """
        return self._public_facing


type AncestryEntityId = tuple[type[Entity], str]


def persistent_id(entity_or_id: Entity | str, /) -> bool:
    """
    Test if an entity ID is persistent.

    See :py:class:`betty.model.NonPersistentId`.
    """
    return not isinstance(
        entity_or_id if isinstance(entity_or_id, str) else entity_or_id.id,
        NonPersistentId,
    )



"""
Provide the Note entity type and utilities.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, final, override

from betty.ancestry.has_links import HasLinks
from betty.ancestry.media_type import HasMediaType
from betty.locale.localizable.gettext import _, ngettext
from betty.locale.localizable.linked_data import dump_linked_data
from betty.locale.localizable.property import LocalizableProperty
from betty.locale.localizable.static.schema import StaticTranslationsSchema
from betty.model import EntityDefinition
from betty.model.association import BidirectionalToZeroOrOne, ToZeroOrOneAssociate
from betty.privacy import HasPrivacy, Privacy, is_public

if TYPE_CHECKING:
    from betty.ancestry.has_notes import HasNotes
    from betty.json.linked_data import JsonLdObject
    from betty.locale.localizable import Localizable, ResolvableLocalizable
    from betty.portable import PortableMapping
    from betty.project import Project


@final
@EntityDefinition(
    "note",
    label=_("Note"),
    label_plural=_("Notes"),
    label_countable=ngettext("{count} note", "{count} notes"),
)
class Note(HasPrivacy, HasLinks, HasMediaType):
    """
    .. plugin:: entity:note.
    """

    text = LocalizableProperty(label=_("Text"))
    """
    The note text.
    """

    entity = BidirectionalToZeroOrOne["Note", "HasNotes"](
        "betty.ancestry.has_notes:HasNotes",
        "notes",
        label=_("Owner"),
        description=_("The entity the note belongs to"),
    )
    """
    The entity the note belongs to.
    """

    def __init__(
        self,
        text: ResolvableLocalizable,
        *,
        id: str | None = None,  # noqa: A002
        entity: ToZeroOrOneAssociate[HasNotes] | None = None,
        privacy: Privacy | None = None,
    ):
        super().__init__(id, privacy=privacy)
        self.text = text
        if entity is not None:
            self.entity = entity

    @override
    @property
    def label(self) -> Localizable:
        return self.text

    @override
    async def dump_linked_data(self, project: Project, /) -> PortableMapping:
        portable = await super().dump_linked_data(project)
        portable["@type"] = "https://schema.org/Thing"
        if is_public(self):
            portable["text"] = dump_linked_data(
                self.text, localizers=await project.public_localizers
            )
        return portable

    @override
    @classmethod
    async def linked_data_schema(cls, project: Project, /) -> JsonLdObject:
        schema = await super().linked_data_schema(project)
        schema.add_property(
            "text",
            StaticTranslationsSchema(),
            False,
        )
        return schema



"""
JSON schemas for the locale API.
"""

from __future__ import annotations

from typing import final

from betty.classtools import Singleton
from betty.json.schema import String


@final
class LocaleSchema(Singleton, String):
    """
    The JSON Schema for locales.
    """

    def __init__(self):
        super().__init__(
            def_name="locale",
            title="Locale",
            description="A BCP 47 locale identifier (https://www.ietf.org/rfc/bcp/bcp47.txt).",
        )



"""
Machine names.
"""

from __future__ import annotations

import re
from typing import TYPE_CHECKING, Self, final, override

from betty.assertion import assert_str
from betty.data import Data, DataDefinition
from betty.exception import HumanFacingException
from betty.locale.localizable.gettext import _
from betty.locale.localizable.markup import Paragraph
from betty.portable import Portable, PortableData
from betty.property import Property

if TYPE_CHECKING:
    from betty.locale.localizable import ResolvableLocalizable

_MACHINE_NAME_DESCRIPTION = _(
    "A machine name is an identifier of at most 250 characters long, made up of lowercase letters, numbers, and/or hyphens (-)."
)
_MACHINE_NAME_PATTERN = re.compile(r"^[a-z0-9\-]{1,250}$")
_MACHINIFY_DISALLOWED_CHARACTER_PATTERN = re.compile(r"[^a-z0-9\-]")
_MACHINIFY_HYPHEN_PATTERN = re.compile(r"-{2,}")


@final
@DataDefinition(label=_("Machine name"))
class MachineName(Portable[str], str, Data):
    """
    A machine name.

    A machine name is a string that meets these criteria:
    - At most 250 characters long.
    - Lowercase letters, numbers, and hyphens (-).
    """

    __slots__ = ()

    @override
    def __new__(cls, machine_name: str, /):
        if _MACHINE_NAME_PATTERN.fullmatch(machine_name) is None:
            raise InvalidMachineName(machine_name)
        return super().__new__(cls, machine_name)

    def __init__(self, machine_name: str, /):
        pass

    @override
    @classmethod
    def load(cls, portable: PortableData, /) -> Self:
        return cls(assert_str()(portable))

    @override
    def dump(self) -> str:
        return self

    @classmethod
    def resolve(cls, machine_name: ResolvableMachineName) -> MachineName:
        """
        Resolve a value to a machine name.
        """
        if isinstance(machine_name, cls):
            return machine_name
        return cls(machine_name)

    @classmethod
    def machinify(cls, source: str, /) -> Self | None:
        """
        Attempt to convert a source string into a valid machine name.
        """
        machine_name = (
            _MACHINIFY_HYPHEN_PATTERN.sub(
                "-", _MACHINIFY_DISALLOWED_CHARACTER_PATTERN.sub("-", source.lower())
            ).strip("-")[:250]
            or None
        )
        if machine_name is None:
            return None
        return cls(machine_name)


type ResolvableMachineName = MachineName | str


@final
class MachineNameProperty(Property):
    """
    A property containing a machine name.
    """

    def __init__(
        self,
        *,
        label: ResolvableLocalizable | None = None,
        description: ResolvableLocalizable | None = None,
    ):
        super().__init__(
            data=MachineName,
            label=_("Name") if label is None else label,
            description=_MACHINE_NAME_DESCRIPTION
            if description is None
            else description,
            resolver=MachineName.resolve,
        )


@final
class InvalidMachineName(HumanFacingException, ValueError):
    """
    Raised when something is not a valid machine name.
    """

    def __init__(self, value: str, /):
        super().__init__(
            Paragraph(
                _('"{value}" is not a valid machine name.').format(value=value),
                _MACHINE_NAME_DESCRIPTION,
            )
        )



"""
Life cycle and resource management.
"""

from __future__ import annotations

from collections.abc import Awaitable, Callable
from typing import TYPE_CHECKING, Any, Self, TypedDict, Unpack, final
from warnings import warn

if TYPE_CHECKING:
    from types import TracebackType


class LifeCycleError(RuntimeError):
    """
    Raised for life cycle related errors.
    """


@final
class NotYetBootstrapped(LifeCycleError):
    """
    Raised if a life cycle was unexpectedly not yet bootstrapped.
    """


@final
class AlreadyBootstrapped(LifeCycleError):
    """
    Raised if a life cycle was unexpectedly already bootstrapped.
    """


@final
class AlreadyShutDown(LifeCycleError):
    """
    Raised if a life cycle was unexpectedly already shut down.
    """


class LifeCycle:
    """
    An object that can bootstrap and shut down.
    """

    def __init__(self, *args: Any, **kwargs: Any):
        self.__bootstrapped = False
        self.__shut_down = False
        super().__init__(*args, **kwargs)

    @final
    @property
    def bootstrapped(self) -> bool:
        """
        Whether the object has been bootstrapped.
        """
        return self.__bootstrapped

    @final
    @property
    def alive(self) -> bool:
        """
        Whether the object is alive, e.g. bootstrapped but not shut down.
        """
        return self.bootstrapped and not self.shut_down

    @final
    @property
    def shut_down(self) -> bool:
        """
        Whether the object has been shut down.
        """
        return self.__shut_down

    @final
    def assert_not_bootstrapped(self) -> None:
        """
        Assert that the object is not yet bootstrapped.
        """
        if self.bootstrapped:
            raise AlreadyBootstrapped(f"{self} was bootstrapped already.")

    @final
    def assert_bootstrapped(self) -> None:
        """
        Assert that the object is bootstrapped.
        """
        if not self.bootstrapped:
            raise NotYetBootstrapped(f"{self} was not bootstrapped yet.")

    @final
    def assert_alive(self) -> None:
        """
        Assert that the object is alive, e.g. bootstrapped but not shut down.
        """
        self.assert_bootstrapped()
        self.assert_not_shut_down()

    @final
    def assert_not_shut_down(self) -> None:
        """
        Assert that the object is not yet shut down.
        """
        if self.shut_down:
            raise AlreadyShutDown(f"{self} was shut down already.")

    async def bootstrap(self) -> None:
        """
        Bootstrap the object.
        """
        self.assert_not_bootstrapped()
        self.__bootstrapped = True

    async def shutdown(self, *, wait: bool = True) -> None:
        """
        Shut the object down.
        """
        self.assert_alive()
        self.__shut_down = True

    @final
    async def __aenter__(self) -> Self:
        await self.bootstrap()
        return self

    @final
    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        await self.shutdown(wait=exc_val is None)

    def __del__(self) -> None:
        if self.alive:
            warn(f"{self} was bootstrapped, but never shut down.", stacklevel=2)


type Bootstrapper = Callable[[], Awaitable[None] | None]
"""
A callback to bootstrap resources.
"""


class ShutdownerKwargs(TypedDict):
    """
    The keyword arguments to a shutdown callback.
    """

    wait: bool
    """
    ``True`` to wait for the component to shut down gracefully, or ``False`` to attempt an immediate forced shutdown.
    """


type Shutdowner = Callable[[Unpack[ShutdownerKwargs]], Awaitable[None] | None]
"""
A callback to shut down resources.
"""



"""
Handle requirements on extensions.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Concatenate, overload

from betty.extension import Extension, ExtensionDefinition
from betty.locale.localizable.gettext import _
from betty.plugin import ResolvableId, resolve_id
from betty.plugin.error import PluginNotFound
from betty.service.requirement import CallableRequirement, Requirement, UnmetRequirement
from betty.service.requirement.project import require_project

if TYPE_CHECKING:
    from collections.abc import Awaitable, Callable

    from betty.project import Project
    from betty.service.level import ServiceLevel


@overload
def require_extension[ExtensionT: Extension](
    extension_id: type[ExtensionT] | ResolvableId[ExtensionDefinition], /
) -> Requirement[ExtensionT]:
    pass


@overload
def require_extension[ExtensionT: Extension, **P, ReturnT](
    extension_id: type[ExtensionT] | ResolvableId[ExtensionDefinition],
    f: Callable[Concatenate[ExtensionT, P], Awaitable[ReturnT] | ReturnT]
    | Callable[Concatenate[Any, ExtensionT, P], Awaitable[ReturnT] | ReturnT],
    /,
) -> CallableRequirement[ExtensionT, P, ReturnT]:
    pass


@overload
def require_extension[ExtensionT: Extension](
    extension_id: type[ExtensionT] | ResolvableId[ExtensionDefinition],
    services: ServiceLevel,
    /,
) -> Awaitable[ExtensionT]:
    pass


def require_extension(extension_id, f=None):
    """
    Decorate a service level callable to require an :py:class:`betty.extension.Extension`.
    """
    requirement = Requirement(
        lambda services, target: _require_extension(services, target, extension_id)
    )
    if f is None:
        return requirement
    return requirement(f)


@require_project
async def _require_extension[ExtensionT: Extension](
    project: Project,
    target: str,
    extension_id: type[ExtensionT] | ResolvableId[ExtensionDefinition],
) -> ExtensionT:
    extensions = await project.plugins.plugins(ExtensionDefinition)
    try:
        extension = extensions[resolve_id(extension_id)]
    except PluginNotFound as error:
        raise UnmetRequirement(error) from error
    project_extensions = await project.extensions
    if extension_id in project_extensions:
        return project_extensions[extension_id]  # ty:ignore[invalid-return-type]
    raise UnmetRequirement(
        _(
            "{target} requires the {extension} extension. Enable it in your project configuration, and try again."
        ).format(target=target, extension=extension.label)
    )



"""
Provide Betty's ancestry genders.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, final, override

from betty.definition.human_facing import CountableHumanFacingDefinition
from betty.locale.localizable.gettext import _, ngettext
from betty.plugin import Plugin, PluginDefinition, PluginTypeDefinition
from betty.plugin.discovery.entry_point import EntryPointDiscovery
from betty.plugin.factory import PluginManufacturer
from betty.service.requirement.project import require_project

if TYPE_CHECKING:
    import builtins


class Gender(Plugin["GenderDefinition"]):
    """
    Define a gender.
    """


@final
@PluginTypeDefinition(
    "gender",
    label=_("Gender"),
    label_plural=_("Genders"),
    label_countable=ngettext("{count} gender", "{count} genders"),
    discovery=[
        EntryPointDiscovery("betty.gender"),
        require_project(
            lambda project: (
                configuration.new_plugin()
                for configuration in project.configuration.genders
            )
        ),
    ],
)
class GenderDefinition(CountableHumanFacingDefinition, PluginDefinition[Gender]):
    """
    .. plugin_type:: gender.

    From `gender <https://en.wikipedia.org/wiki/Gender>`_ on Wikipedia:

        Gender includes the social, psychological, cultural and behavioral aspects of being a man, woman, or other gender
        identity.
    """


@final
class GenderManufacturer(PluginManufacturer[GenderDefinition, Gender]):
    """
    The gender manufacturer.
    """

    @override
    @classmethod
    def type(cls) -> builtins.type[GenderDefinition]:
        return GenderDefinition



"""
JSON schemas for the model API.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

from betty.json.schema import Array, Null, OneOf, String

if TYPE_CHECKING:
    from betty.locale.localizable import ResolvableLocalizable


class ToZeroOrOneSchema(OneOf):
    """
    A schema for a to-zero-or-one entity association.
    """

    def __init__(
        self,
        *,
        title: ResolvableLocalizable | None = None,
        description: ResolvableLocalizable | None = None,
    ):
        super().__init__(
            String(
                title=title or "Optional associate entity",
                description=description
                or "An optional reference to an associate entity's JSON resource",
                format=String.Format.URI,
            ),
            Null(),
        )


class ToOneSchema(String):
    """
    A schema for a to-one entity association.
    """

    def __init__(
        self,
        *,
        title: ResolvableLocalizable | None = None,
        description: ResolvableLocalizable | None = None,
    ):
        super().__init__(
            title=title or "Associate entity",
            description=description
            or "A reference to an associate entity's JSON resource",
            format=String.Format.URI,
        )


class ToManySchema(Array):
    """
    A schema for a to-many entity association.
    """

    def __init__(
        self,
        *,
        title: ResolvableLocalizable | None = None,
        description: ResolvableLocalizable | None = None,
    ):
        super().__init__(
            ToOneSchema(),
            title=title or "Associate entities",
            description=description
            or "References to associate entities' JSON resources",
        )
